local CrossEntropyCriterion, Criterion = torch.class('nn.SepCrossEntropyCriterion', 'nn.Criterion')

function CrossEntropyCriterion:__init(nb, weights)
   Criterion.__init(self)
   self.nb = nb or 24
   self.lsm = nn.LogSoftMax()
   self.nll = nn.ClassNLLCriterion(weights)
end

function CrossEntropyCriterion:updateOutput(input, target)
   --input = input:squeeze()
   target = target:squeeze()
   local num_cls = input:size(2)/self.nb
   local sinput = input:view(-1, self.nb)
   local output = 0
   for i = 1, input:size(1) do
      local cl = math.floor((target[i]-1)/self.nb)-- + 1
      local ang = (target[i]-1)%self.nb + 1
      local linput = input[i]:narrow(1,cl*self.nb+1,self.nb)

      self.lsm:updateOutput(linput)
      self.nll:updateOutput(self.lsm.output, ang)
      output = output + self.nll.output
   end
   self.output = output/input:size(1)
   return self.output
end

function CrossEntropyCriterion:updateGradInput(input, target)
   local size = input:size()
   target = target:squeeze()
   local num_cls = input:size(2)/self.nb
   local sinput = input:view(-1, self.nb)
   self.gradInput:resizeAs(input):zero()
   for i = 1, input:size(1) do
      local cl = math.floor((target[i]-1)/self.nb)-- + 1
      local ang = (target[i]-1)%self.nb + 1
      local linput = input[i]:narrow(1, cl*self.nb+1, self.nb)
   
      self.lsm:updateOutput(linput)
      self.nll:updateOutput(self.lsm.output, ang)
      self.nll:updateGradInput(self.lsm.output, ang)
      self.lsm:updateGradInput(linput, self.nll.gradInput)

      self.gradInput[i]:narrow(1,cl*self.nb+1,self.nb):copy(self.lsm.gradInput)

   end

   self.gradInput:div(input:size(1))

   return self.gradInput
end

--return nn.CrossEntropyCriterion

local function criterionJacobianTest1D(cri, input, target)
  local precision = 1e-5
  local eps = 1e-6
  local _ = cri:forward(input, target)
  local dfdx = cri:backward(input, target)
  -- for each input perturbation, do central difference
  local centraldiff_dfdx = torch.Tensor():resizeAs(dfdx)
  local input_s = input:storage()
  local centraldiff_dfdx_s = centraldiff_dfdx:storage()
  for i=1,input:nElement() do
    -- f(xi + h)
    input_s[i] = input_s[i] + eps
    local fx1 = cri:forward(input, target)
    -- f(xi - h)
    input_s[i] = input_s[i] - 2*eps
    local fx2 = cri:forward(input, target)
    -- f'(xi) = (f(xi + h) - f(xi - h)) / 2h
    local cdfx = (fx1 - fx2) / (2*eps)
    -- store f' in appropriate place
    centraldiff_dfdx_s[i] = cdfx
    -- reset input[i]
    input_s[i] = input_s[i] + eps
  end

  -- compare centraldiff_dfdx with :backward()
  local err = (centraldiff_dfdx - dfdx):abs():max()
  --mytester:assertlt(err, precision, 'error in difference between central difference and :backward')
  print(err)
  assert(err < precision, 'error in difference between central difference and :backward')
end



function CrossEntropyCriterion:test(bs)
  local numLabels = self.nb*12--math.random(5,10)
  bs = bs or 1
  local input = torch.rand(bs,numLabels)
  local target = torch.Tensor(bs):random(1,numLabels)

  criterionJacobianTest1D(self, input, target)
end



