local Criterion, parent = torch.class('nnf.SoftMaxWithViewCriterion', 'nn.Criterion')

local tol = 1e-12
local ttype = torch.FloatTensor

function Criterion:__init(bandwidth,sigma,period)
  parent.__init(self)
  self.softmax = nn.SoftMax()
  self.bandwidth = bandwidth or 5
  self.sigma = sigma or 3
  self.period = period or 360

  local ws = 0
  for k = -self.bandwidth, self.bandwidth do
    ws = ws + math.exp(-math.abs(k)/self.sigma)
  end
  self.weight_sum = ws
  self.with_bg = false
end

function Criterion:updateOutput(input, target)
  local bandwidth = self.bandwidth
  local sigma = self.sigma
  local period = self.period
  local dim = input:size(2)

  local num = target:size(1)
  local loss = 0
  local add_bg = 0
  if self.with_bg then
    add_bg = 1
  end

  self._output = self._output or ttype()
  local o = self.softmax:updateOutput(input)
  self._output:resize(o:size()):copy(o)

  self._target = self._target or ttype()
  self._target:resize(target:size(1)):copy(target)
  for i = 1, num do
    local label_value = self._target[i] - 2
    if label_value > -1 then -- skip bg
      local cls_idx = math.floor((label_value)/period)--+2
      local inp = self._output:select(1,i):narrow(1,(cls_idx)*period+1+add_bg, period)
      local prob_cls_sum = inp:sum()
      inp:div(prob_cls_sum)
      local prob_data = inp:data()

      local view_label = (label_value)%period -- 0 -> 359
      local tmp_loss = 0
      for k = -bandwidth, bandwidth do
        local view_k = (view_label + k)%period
        tmp_loss = tmp_loss - math.exp(-math.abs(k)/sigma) * math.log(math.max(prob_data[view_k ],tol))
      end
      loss = loss + tmp_loss
    end
  end
  self.output = loss/num
  return self.output
end

function Criterion:updateGradInput(input, target)
  local bandwidth = self.bandwidth
  local sigma = self.sigma
  local period = self.period
  local dim = input:size(2)

  local add_bg = 0
  if self.with_bg then
    add_bg = 1
  end

  local num = target:size(1)

  self._gradInput = self._gradInput or ttype()
  self._gradInput:resize(input:size()):zero()
  self.gradInput:resizeAs(input)

  for i = 1, num do
    local label_value = self._target[i] - 2
    if label_value > -1 then -- skip bg
      local cls_idx = math.floor((label_value)/period)--+2
      local inp = self._output:select(1,i):narrow(1,(cls_idx)*period+1+add_bg, period)
      local gi = self._gradInput:select(1,i):narrow(1,(cls_idx)*period+1+add_bg,period)
      gi:copy(inp):mul(self.weight_sum)
      local gi_data = gi:data()

      local view_label = (label_value)%period -- 0 -> 359
      for k = -bandwidth, bandwidth do
        local view_k = (view_label + k)%period
        gi_data[view_k] = gi_data[view_k] - math.exp(-math.abs(k)/sigma)
      end
    end
  end
  self.gradInput:copy(self._gradInput)

  local loss_weight = 1
  self.gradInput:mul(loss_weight/num)

  return self.gradInput
end

local function criterionJacobianTest1D(cri, input, target)
  local precision = 1e-5
  local eps = 1e-6
  local _ = cri:forward(input, target)
  local dfdx = cri:backward(input, target)
  -- for each input perturbation, do central difference
  local centraldiff_dfdx = torch.Tensor():resizeAs(dfdx)
  local input_s = input:storage()
  local centraldiff_dfdx_s = centraldiff_dfdx:storage()
  for i=1,input:nElement() do
    -- f(xi + h)
    input_s[i] = input_s[i] + eps
    local fx1 = cri:forward(input, target)
    -- f(xi - h)
    input_s[i] = input_s[i] - 2*eps
    local fx2 = cri:forward(input, target)
    -- f'(xi) = (f(xi + h) - f(xi - h)) / 2h
    local cdfx = (fx1 - fx2) / (2*eps)
    -- store f' in appropriate place
    centraldiff_dfdx_s[i] = cdfx
    -- reset input[i]
    input_s[i] = input_s[i] + eps
  end

  -- compare centraldiff_dfdx with :backward()
  local err = (centraldiff_dfdx - dfdx):abs():max()
  --mytester:assertlt(err, precision, 'error in difference between central difference and :backward')
  print(err)
  assert(err < precision, 'error in difference between central difference and :backward')
end



function Criterion:test(bs)
  local numLabels = 360*12+1--math.random(5,10)
  bs = bs or 1
  local input = torch.rand(bs,numLabels)
  local target = torch.Tensor(bs):random(1,numLabels)

  criterionJacobianTest1D(self, input, target)
end

function Criterion:testspeed(bs)
  local numLabels = 360*12+1--math.random(5,10)
  bs = bs or 1
  local input = torch.rand(bs,numLabels)
  local target = torch.Tensor(bs):random(1,numLabels)

  local out = self:forward(input, target)
  local g = self:backward(input, target)
end
