require 'cudnn'
require 'inn'

-- similar to nn.Module.apply
-- goes over a net and recursively replaces modules
-- using callback function
local function replace(self, callback)
  local out = callback(self)
  if self.modules then
    for i, module in ipairs(self.modules) do
      self.modules[i] = replace(module, callback)
    end
  end
  return out
end

local function BNtoFixed(net, ip)
  return replace(
  net,
  function(x)
    if torch.typename(x):find'SpatialBatchNormalization' then
      local no = x.running_mean:numel()
      local y = nnf.SpatialConstDiagonal(no, ip):type(x.running_mean:type())
      if x.running_var then
        x.running_std = x.running_var:add(x.eps):pow(-0.5)
      end
      y.a:copy(x.running_std)
      y.b:add(-1,x.running_mean):cmul(x.running_std)
      if x.affine then
        y.a:cmul(x.weight)
        y.b:cmul(x.weight):add(x.bias)
      end
      return y
    else
      return x
    end
  end
  )
end

local function createRegressionOutput(opt)
  local pose_n = nn.ConcatTable()
  for i = 1, 12 do
    local s = nn.Sequential()
    local l = nn.Linear(4096,opt.manif_size)
    l.weight:normal(0,0.01)
    l.bias:zero()
    s:add(l)
    if opt.normalize then
      s:add(nn.Normalize(2))
    end
    pose_n:add(s)
  end
  return pose_n
end

local function createModel(opt)
  local model
  local num_output = 21
  if opt.ptype == 'discrete' or opt.ptype == 'discrete-altnorm' then
    local add_bg = opt.joint_trained == true and 1 or 0
    num_output = 12*opt.ang_res + add_bg
  elseif opt.ptype == 'mixed' then
    local add_bg = opt.joint_trained == true and 1 or 0
    num_output = 12 + add_bg
  elseif opt.ptype == 'rendercnn' then
    local add_bg = opt.joint_trained == true and 1 or 0
    num_output = 12*360 + add_bg
  elseif opt.ptype == 'nopose' then
    num_output = 12 + 1
  end
  if opt.model == 'resnet18' then
    local base_model = torch.load('models/resnet-18.t7')
    base_model:remove() -- linear 512->1000
    base_model:remove() -- view
    base_model:remove() -- averagepooling 7x7,1,1

    model = nn.Sequential()
    local features = nn.Sequential()
    for i=1,7 do
      features:add(base_model:get(i))
    end

    local classifier = nn.Sequential()
    classifier:add(base_model:get(8))
    classifier:add(nn.SpatialAveragePooling(7,7,1,1))
    classifier:add(nn.View(-1):setNumInputDims(3))
    local linear = nn.Linear(512,num_output)
    linear.bias:zero()
    classifier:add(linear)

    local prl = nn.ParallelTable()
    prl:add(features)
    prl:add(nn.Identity())

    local ROIPooling = inn.ROIPooling(14,14):setSpatialScale(1/16)

    model:add(prl)
    model:add(ROIPooling)
    model:add(classifier)

    if opt.bn_mode == 'buggy' then
      BNtoFixed(model, true)
    elseif opt.bn_mode == 'fix' then
      model:apply(function(x)
        if torch.typename(x):find('BatchNormalization') then
          x:freezeStatistics(true)
        end
      end)
    end

  elseif opt.model == 'resnet50' then
    local base_model = torch.load('models/resnet-50.t7')
    base_model:remove() -- linear 2048->1000
    base_model:remove() -- view
    base_model:remove() -- averagepooling 7x7,1,1

    model = nn.Sequential()
    local features = nn.Sequential()
    for i=1,7 do
      features:add(base_model:get(i))
    end

    local classifier = nn.Sequential()
    classifier:add(base_model:get(8))
    classifier:add(nn.SpatialAveragePooling(7,7,1,1))
    classifier:add(nn.View(-1):setNumInputDims(3))
    local linear = nn.Linear(2048,num_output)
    linear.bias:zero()
    classifier:add(linear)

    local prl = nn.ParallelTable()
    prl:add(features)
    prl:add(nn.Identity())

    local ROIPooling = inn.ROIPooling(14,14):setSpatialScale(1/16)

    model:add(prl)
    model:add(ROIPooling)
    model:add(classifier)

    if opt.bn_mode == 'buggy' then
      BNtoFixed(model, true)
    elseif opt.bn_mode == 'fix' then
      model:apply(function(x)
        if torch.typename(x):find('BatchNormalization') then
          x:freezeStatistics(true)
        end
      end)
    end


  elseif opt.model == 'alexnet' then
    model = torch.load('models/alexnet_frcnn.t7')
    model:get(4):remove() -- remove 1000 layer

    local linear = nn.Linear(4096, num_output)
    linear.weight:normal(0,0.01)
    linear.bias:zero()

    local top
    if opt.ptype == 'discrete' or opt.ptype == 'rendercnn' or opt.ptype == 'nopose' or opt.ptype == 'discrete-altnorm' then
      top = linear
    elseif opt.ptype == 'mixed' then
      top = nn.ConcatTable()
      local seq = nn.Sequential()
      local pose_n = createRegressionOutput(opt)
      if opt.ang_place == 'fc7' then
        seq:add(linear)
        top:add(seq)
        if opt.no_bp_ang then
          local s = nn.Sequential()
          s:insert(nnf.BlockGradient())
          s:insert(pose_n)
          top:add(s)
        else
          top:add(pose_n)
        end
      elseif opt.ang_place == 'pool5' then
        local cl = model:get(4)
        model:remove() -- remove entire classifier
        local br1 = cl:clone()
        br1:add(linear)
        local br2 = cl:clone()
        br2:add(pose_n)
        if opt.no_bp_ang then
          br2:insert(nnf.BlockGradient(),1)
        end
        top:add(br1)
        top:add(br2)
      elseif opt.ang_place == 'conv4' then
        local cl = model:get(4)
        model:remove() -- remove entire classifier
        local last_conv = model:get(1):get(1):get(13)
        model:get(1):get(1):remove() -- relu
        model:get(1):get(1):remove() -- last conv
        -- change ROIPooling
        model:get(2).W = 12
        model:get(2).H = 12
        model:remove(3) -- view

        cl:insert(last_conv,1)
        cl:insert(cudnn.ReLU(true),2)
        cl:insert(cudnn.SpatialMaxPooling(3,3,2,2,1,1),3)
        cl:insert(nn.View(-1):setNumInputDims(3),4)

        local br1 = cl:clone()
        br1:add(linear)
        local br2 = cl:clone()
        br2:add(pose_n)
        if opt.no_bp_ang then
          br2:insert(nnf.BlockGradient(),1)
        end
        top:add(br1)
        top:add(br2)

      end
    elseif opt.ptype == 'regression' then
      top = nn.Sequential()
      local pose_n = createRegressionOutput(opt)
      if opt.ang_place == 'fc7' then
        if opt.no_bp_ang then
          local s = nn.Sequential()
          s:insert(nnf.BlockGradient())
          s:insert(pose_n)
          top:add(s)
        else
          top:add(pose_n)
        end
      elseif opt.ang_place == 'pool5' then
        local cl = model:get(4)
        model:remove() -- remove entire classifier
        local br2 = cl:clone()
        br2:add(pose_n)
        if opt.no_bp_ang then
          br2:insert(nnf.BlockGradient(),1)
        end
        top:add(br2)
      end
    end

    --model:get(4):add(top)
    model:add(top)
  elseif opt.model == 'vgg' then
    local base_model = torch.load('models/vgg.t7')
    --cudnn.convert(base_model, cudnn)
    base_model:remove() -- softmax
    base_model:remove() -- linear 4096 -> 1000

    model = nn.Sequential()
    local features = nn.Sequential()
    for i=1,30 do
      features:add(base_model:get(i))
    end

    local classifier = nn.Sequential()
    for i=33,38 do
      classifier:add(base_model:get(i))
    end

    local linear = nn.Linear(4096, num_output)
    linear.weight:normal(0, 0.01)
    linear.bias:zero()
    classifier:add(linear)

    local prl = nn.ParallelTable()
    prl:add(features)
    prl:add(nn.Identity())

    local ROIPooling = inn.ROIPooling(7,7):setSpatialScale(1/16)

    model:add(prl)
    model:add(ROIPooling)
    model:add(nn.View(-1):setNumInputDims(3))
    model:add(classifier)

  end
  return model
end

return createModel
