require 'nnf'
require 'cunn'

cmd = torch.CmdLine()
cmd:text('Example on how to train/test a RCNN based object detector on Pascal')
cmd:text('')
cmd:text('Options:')
cmd:option('-name',      'frcnn-pose',   'base name')
cmd:option('-dataset',   'voc2012train', 'train dataset')
cmd:option('-proposals', 'deepmask',     'proposals for training')
cmd:option('-testproposals', 'deepmask',     'proposals for test')
cmd:option('-lr',        1e-3,           'learning rate')
cmd:option('-num_iter',  40000,          'number of iterations')
cmd:option('-disp_iter', 100,            'display every n iterations')
cmd:option('-lr_step',   30000,          'step for reducing the learning rate')
cmd:option('-save_step', 10000,          'step for saving the model')
cmd:option('-gpu',       1,              'gpu to use (0 for cpu mode)')
cmd:option('-ptype',     'discrete',     'pose type')
cmd:option('-crit',      'smoothl1',     'criterion for the regression')
cmd:option('-reg_weight', 1,             'Weight for the regression criterion')
cmd:option('-normalize', false,          'l2-normalize pose output')
cmd:option('-manif_size',2,              'size of the regression manifold')
cmd:option('-ang_place', 'fc7',          'Position of the regression branch')
cmd:option('-no_bp_ang', false,          'Block propagation from regression branch')
cmd:option('-ang_res',   24,             'Angle resolution')
cmd:option('-model',     'resnet18',     'model to use')
cmd:option('-bn_mode',   'buggy',        'BN mode to use')
cmd:option('-only_test', false,          'only for evaluation')
cmd:option('-seed',      1,              'fix random seed (if ~= 0)')
cmd:option('-imgs_per_batch',     -1,    'number of images per batch')
cmd:option('-batch_size', 128,           'batch size')
cmd:option('-finetune',  false,          'finetuning from trained net')
cmd:option('-bp_lo',     0,              'batch provider bg low threshold')
cmd:option('-bp_hi',     0.5,            'batch provider bg high threshold')
cmd:option('-hard_gt',   true,           'Use hard samples')
cmd:option('-tflip',     true,           'Use training flip')
cmd:option('-synth_data',-1,             'Use synthetic data')
cmd:option('-sep_class', false,          'Separate classifier')
cmd:option('-joint_trained', true,       'Joint training the pose and the classifier (false = do not add +1 for bg)')
cmd:option('-subset',    -1,             'number of ImageNet images to use per class')
cmd:option('-numthreads',6,              'number of threads')

opt = cmd:parse(arg or {})

if not opt.sep_class and not opt.joint_trained then
  error('Cant use sep_class=false with joint_trained=false')
end

exp_name = cmd:string(opt.name, opt, {name=true, gpu=true, numthreads=true, only_test=true,
                                      ptype=false, ang_res=false, tflip=false, finetune=true})

rundir = 'cachedir/'..exp_name
paths.mkdir(rundir)

if opt.only_test then
  cmd:log(paths.concat(rundir,'log_test'), opt)
  cmd:addTime('Fast-RCNN')
elseif not opt.finetune then
  cmd:log(paths.concat(rundir,'log'), opt)
  cmd:addTime('Fast-RCNN')
else
  cmd:log(paths.concat(rundir,'log_finetune'), opt)
  cmd:addTime('Fast-RCNN')
end

local tensor_type
if opt.gpu > 0 then
  require 'cunn'
  cutorch.setDevice(opt.gpu)
  tensor_type = 'torch.CudaTensor'
  print('Using GPU mode on device '..opt.gpu)
else
  require 'nn'
  tensor_type = 'torch.FloatTensor'
  print('Using CPU mode')
end

if opt.seed ~= 0 then
  torch.manualSeed(opt.seed)
  if opt.gpu > 0 then
    cutorch.manualSeed(opt.seed)
  end
  print('Using fixed seed: '..opt.seed)
end

torch.setnumthreads(opt.numthreads)

--------------------------------------------------------------------------------
-- define model and criterion
--------------------------------------------------------------------------------
--local createModel = paths.dofile('../models/frcnn_alexnet.lua')
local createModel = paths.dofile('create_model.lua')
if not opt.only_test or opt.finetune then
  model = createModel(opt)
end
if opt.only_test or opt.finetune then
  model = torch.load(('cachedir/%s/model.t7'):format(exp_name))
end

if opt.ptype == 'discrete' or opt.ptype == 'nopose' or opt.ptype == 'sep-discrete' then
  criterion = nn.CrossEntropyCriterion()
elseif opt.ptype == 'mixed' then
  criterion = nn.ParallelCriterion()
  criterion:add(nn.CrossEntropyCriterion())
  local crit
  if opt.crit == 'smoothl1' then
    crit = nnf.SelectCriterion(nn.SmoothL1Criterion(false))
  elseif opt.crit == 'mse' then
    crit = nnf.SelectCriterion(nn.MSECriterion(false))
  end
  criterion:add(crit, opt.reg_weight)
elseif opt.ptype == 'rendercnn' then
  criterion = nnf.SoftMaxWithViewCriterion()
elseif opt.ptype == 'regression' then
  if opt.crit == 'smoothl1' then
    criterion = nnf.SelectCriterion(nn.SmoothL1Criterion(false))
  elseif opt.crit == 'mse' then
    criterion = nnf.SelectCriterion(nn.MSECriterion(false))
  end
elseif opt.ptype == 'discrete-altnorm' then
  dofile 'SepCrossEntropyCriterion.lua'
  criterion = nn.SepCrossEntropyCriterion()
end

model:type(tensor_type)
criterion:type(tensor_type)

print('Model:')
print(model)
print('Criterion:')
print(criterion)

print('Optimizing training for memory')
local optnet = require 'optnet'
optnet.removeOptimization(model)
optnet.optimizeMemory(model,{torch.CudaTensor(1,3,224,224):uniform(), torch.CudaTensor{{1,1,1,50,50}}},{mode='training'})

-- define the transformations to do in the image before
-- passing it to the network

local image_transformer
do
  local mean_pix = {
    alexnet = {102.9801,115.9465,122.7717},
    vgg = {103.939, 116.779, 123.68},
    resnet18 = {0.485, 0.456, 0.406},
  }
  local raw_scale = {
    alexnet = 255,
    vgg = 255,
    resnet18 = 1,
  }
  local scale = {
    alexnet = {1,1,1},
    vgg = {1,1,1},
    resnet18 = {0.229, 0.224, 0.225},
  }
  local swap = {
    alexnet = {3,2,1},
    vgg = {3,2,1},
    resnet18 = nil,
  }

  image_transformer= nnf.ImageTransformer{
    mean_pix = mean_pix[opt.model],
    raw_scale = raw_scale[opt.model],
    swap = swap[opt.model],
    scale = scale[opt.model],
  }
end

print(image_transformer)
-- [[
local cls = {'aeroplane','bicycle','boat','bottle','bus','car',
              'chair','diningtable','motorbike',
              'sofa','train','tvmonitor'}
--]]
--------------------------------------------------------------------------------
-- define data for training
--------------------------------------------------------------------------------

-- this class holds all the necessary informationn regarding the dataset
if opt.dataset == 'voc2012train' then
  local roidbfile
  if opt.proposals == 'deepmask' then
    roidbfile = 'data/MAD-DM2k/voc_2012_train.t7'
  elseif opt.proposals == 'selectivesearch' then
    roidbfile = 'data/selective_search_data/voc_2012_train.mat'
  elseif opt.proposals == 'DMSS' then
    roidbfile = 'data/MAD-DM2k/voc_2012_train.t7'
  else
    error('Unknown proposal '.. opt.proposals)
  end
  ds = nnf.DataSetPascal{
    image_set='train',
    datadir='datasets/VOCdevkit',
    dataset_name='Pascal3D',
    roidbfile=roidbfile,
    year=2012,
    classes=cls,
    with_hard_samples=opt.hard_gt
  }
  
  ds:loadROIDB()
  if opt.proposals == 'DMSS' then
    print('Concatenating ROIs from selective search to DeepMask')
    local roidb = ds.roidb
    ds.roidb = nil
    ds.roidbfile = 'data/selective_search_data/voc_2012_train.mat'
    ds:loadROIDB()
    for i = 1, ds:size() do
      ds.roidb[i] = torch.cat(roidb[i], ds.roidb[i], 1)
    end
  end
elseif opt.dataset == 'P3Dtrain' then
  local roidbfile
  if opt.proposals == 'deepmask' then
    --roidbfile = 'data/MAD-DM2k/voc_2012_train.t7'
    error('Dont have proposals for this dataset')
  elseif opt.proposals == 'selectivesearch' then
    roidbfile = 'datasets/VOCdevkit/Pascal3D/imagenet_pascal_train.mat'
  elseif opt.proposals == 'DMSS' then
    roidbfile = 'datasets/VOCdevkit/Pascal3D/imagenet_pascal_train.mat'
    --roidbfile = 'data/MAD-DM2k/voc_2012_train.t7'
  else
    error('Unknown proposal '.. opt.proposals)
  end

  ds = nnf.DataSetPascal{
    image_set='imagenet_pascal_train',
    datadir='datasets/VOCdevkit',
    dataset_name='Pascal3D',
    roidbfile=roidbfile,
    year=2012,
    classes=cls,
    with_hard_samples=opt.hard_gt
  }

  ds:loadROIDB()
  if opt.proposals == 'DMSS' then
    print('Concatenating ROIs from selective search to DeepMask')
    local roidb = ds.roidb
    ds.roidb = nil
    ds.roidbfile = 'data/MAD-DM2k/voc_2012_train.t7'
    ds:loadROIDB()
    for i = 1, ds:size() do
      if ds.roidb[i]:numel() > 0 then
        ds.roidb[i] = torch.cat(roidb[i], ds.roidb[i], 1)
      else
        ds.roidb[i] = roidb[i]
      end
    end
  end
  if opt.subset ~= -1 then
    print('Keeping only a subset of ImageNet with '.. opt.subset..' images')
    -- to avoid performance influence due to rng
    local rngstate = torch.getRNGState()
    local ds_info = {}
    local is_pascal = {}
    for i = 1, ds:size() do
      local imn = ds.img_ids[i]
      if imn:sub(1,1) == 'n' then
        local info2 = torch.zeros(ds.num_classes+1):totable()
        info2[ds.num_classes+1] = i
        local anno = ds:getAnnotation(i)
        for j = 1, #anno.object do
          local cl = ds.class_to_id[anno.object[j].name] 
          info2[cl] = info2[cl] + 1
        end
        table.insert(ds_info,info2)
      else
        table.insert(is_pascal,i)
      end
    end
    ds_info = torch.Tensor(ds_info)

    local rp = torch.randperm(ds_info:size(1)):long()
    local shinfo = ds_info:index(1,rp)

    local cls_count = torch.zeros(ds.num_classes)
    local cls_count_cp = cls_count:clone()
    local num_per_class = opt.subset
    local added = {}
    for im = 1, shinfo:size(1) do
      local dd = shinfo[im]
      cls_count_cp:copy(cls_count):add(dd[{{1, ds.num_classes}}])
      if cls_count_cp:le(num_per_class):all() then
        table.insert(added, dd[-1])
        cls_count:copy(cls_count_cp)
      end
    end

    is_pascal = torch.LongTensor(is_pascal)
    local selected_imagenet = torch.LongTensor(added)
    local selected_imgs = torch.cat(selected_imagenet,is_pascal,1)
    selected_imgs = torch.sort(selected_imgs, 1) -- ascending

    local tmp_img_ids = {}
    local tmp_roidb = {}

    for i = 1,selected_imgs:size(1) do
      table.insert(tmp_img_ids, ds.img_ids[selected_imgs[i] ])
      table.insert(tmp_roidb, ds.roidb[selected_imgs[i] ])
    end
    ds.num_imgs = #ds.img_ids
    ds.img_ids = tmp_img_ids
    ds.roidb = tmp_roidb

    -- put rgn state as it should be
    torch.setRNGState(rngstate)
  end
end

ds.save_objs = true
ds:loadROIDB()

print('Removing images without positives')
local dont_keep = {}
for i = 1, ds:size() do
  local _, gt = ds:getGTBoxes(i)
  if #gt == 0 then
    table.insert(dont_keep, i)
  end
end
print('Found '..#dont_keep..' images without positives')
for i = #dont_keep, 1, -1 do
  table.remove(ds.img_ids, dont_keep[i])
  table.remove(ds.roidb, dont_keep[i])
end
ds.num_imgs = #ds.img_ids


-- add flipped images to dataset
--for i = 1, ds.num_imgs do
--  table.insert(ds.img_ids, ds.img_ids[i])
--end

print('DataSet Training:')
print(ds)
--------------------------------------------------------------------------------
-- define feature providers
--------------------------------------------------------------------------------

local crop_size = 224

-- the feature provider extract the features for a given image + bounding box
fp = nnf.FRCNN{
  scale={600},
  max_size=1000,
  image_transformer=image_transformer,
}
-- different frameworks can behave differently during training and testing
fp:training()

print('Feature Provider:')
print(fp)

--------------------------------------------------------------------------------
-- define batch providers
--------------------------------------------------------------------------------

local fg_fraction = 0.25
local imgs_per_batch = 2
local add_bg = 1
if not opt.joint_trained then
  fg_fraction = 1
  imgs_per_batch = 8
  add_bg = 0
end

if opt.imgs_per_batch ~= -1 then
  imgs_per_batch = opt.imgs_per_batch
end

bp = nnf.BatchProviderIC{
  dataset=ds,
  feat_provider=fp,
  bg_threshold={opt.bp_lo, opt.bp_hi},
  batch_size=opt.batch_size,
  do_flip=opt.tflip,
  fg_fraction=fg_fraction,
  imgs_per_batch=imgs_per_batch,
}

local function createWindowAngleBasic(rec,i,j,is_bg)
  local label = is_bg == true and 0+1 or rec.label[j]+1
  local ang
  if is_bg == false and rec.objects[rec.correspondance[j] ] then
    if rec.objects[rec.correspondance[j] ].viewpoint.distance == '0' then
      ang = rec.objects[rec.correspondance[j] ].viewpoint.azimuth_coarse
    else
      ang = rec.objects[rec.correspondance[j] ].viewpoint.azimuth
    end
  else
    ang = 0
  end
  local window = {i,rec.boxes[j][1],rec.boxes[j][2],
                    rec.boxes[j][3],rec.boxes[j][4],
                    label,ang}
  return window
end

local function createWindowAngleDiscrete(rec,i,j,is_bg)
  --local label = is_bg == true and 0+1 or rec.label[j]+1
  local ang
  if is_bg == false and rec.objects[rec.correspondance[j] ] then
    if rec.objects[rec.correspondance[j] ].viewpoint.distance == '0' then
      ang = rec.objects[rec.correspondance[j] ].viewpoint.azimuth_coarse
    else
      ang = rec.objects[rec.correspondance[j] ].viewpoint.azimuth
    end
  else
    ang = 0
  end
  local nPoses = opt.ang_res
  local divAng = 360/nPoses
  local label
  if is_bg then
    label = 1
  else
    label = add_bg + (rec.label[j]-1)*nPoses + math.fmod(math.ceil((ang-divAng/2)/divAng),nPoses)+1
  end
  
  local window = {i,rec.boxes[j][1],rec.boxes[j][2],
                    rec.boxes[j][3],rec.boxes[j][4],
                    label}
  return window
end

--bp:setCreateWindow(createWindowAngleDiscrete)
bp:setCreateWindow(createWindowAngleBasic)

local function flip_angle(x)
  return (-x)%360
end

if opt.ptype == 'discrete' or opt.ptype == 'discrete-altnorm' then
  bp.setLabel = function(self, dd, do_flip)
    local nPoses = opt.ang_res
    local divAng = 360/nPoses

    local tt = dd[{{5,6}}]
    if tt[1] == 1 then
      return {1}
    else
      local ang = tt[2]
      if do_flip == 1 then
        ang = flip_angle(ang)
      end
      return {add_bg + (tt[1]-2)*nPoses + math.fmod(math.ceil((ang-divAng/2)/divAng),nPoses)+1}
    end
  end
elseif opt.ptype == 'mixed' then
  -- [[  
  bp.setLabel = function(self, dd, do_flip)
    local nPoses = opt.ang_res
    local divAng = 360/nPoses

    local tt = dd[{{5,6}}]
    if tt[1] == 1 then
      return {1, 0}
    else
      local ang = tt[2]
      if do_flip == 1 then
        ang = flip_angle(ang)
      end
      return {tt[1], ang}
    end
  end
  --]]
  if opt.manif_size == 2 then
    bp.prepareTarget = function(self, labels)
      local res = {}
      local t = torch.FloatTensor(labels)
      res[1] = t:select(2,1)
      res[2] = {}
      res[2][1] = t:select(2,1)
      local ang = torch.FloatTensor(t:size(1),2)
      ang:select(2,1):cos(t:select(2,2)*(math.pi/180))
      ang:select(2,2):sin(t:select(2,2)*(math.pi/180))
      res[2][2] = ang
      return res
    end
  else
    bp.prepareTarget = function(self, labels)
      local res = {}
      local t = torch.FloatTensor(labels)
      res[1] = t:select(2,1)
      res[2] = {}
      res[2][1] = t:select(2,1)
      local ang = torch.FloatTensor(t:size(1),opt.manif_size)
      --local a_i = torch.linspace(0,90,opt.manif_size)
      local a_i = torch.linspace(-60,60,opt.manif_size)
      for i = 1, opt.manif_size do
        ang:select(2,i):cos((t:select(2,2)-a_i[i])*(math.pi/180))
      end
      --ang:select(2,2):sin(t:select(2,2)*(math.pi/180))
      res[2][2] = ang
      return res
    end
  end
elseif opt.ptype == 'rendercnn' then
  bp.setLabel = function(self, dd, do_flip)
    local tt = dd[{{5,6}}]
    if tt[1] == 1 then
      return {1}
    else
      local ang = tt[2]
      if do_flip == 1 then
        ang = flip_angle(ang)
      end
      return {add_bg + (tt[1]-2)*360 + math.floor(ang)+1}
    end
  end
elseif opt.ptype == 'regression' then
  bp.setLabel = function(self, dd, do_flip)
    local nPoses = opt.ang_res
    local divAng = 360/nPoses

    local tt = dd[{{5,6}}]
    if tt[1] == 1 then
      return {1, 0}
    else
      local ang = tt[2]
      if do_flip == 1 then
        ang = flip_angle(ang)
      end
      return {tt[1], ang}
    end
  end
  --]]
  if opt.manif_size == 2 then
    bp.prepareTarget = function(self, labels)
      local res = {}
      local t = torch.FloatTensor(labels)
      res[1] = t:select(2,1)
      local ang = torch.FloatTensor(t:size(1),2)
      ang:select(2,1):cos(t:select(2,2)*(math.pi/180))
      ang:select(2,2):sin(t:select(2,2)*(math.pi/180))
      res[2] = ang
      return res
    end
  else
    bp.prepareTarget = function(self, labels)
      local res = {}
      local t = torch.FloatTensor(labels)
      res[1] = t:select(2,1)
      local ang = torch.FloatTensor(t:size(1),opt.manif_size)
      --local a_i = torch.linspace(0,90,opt.manif_size)
      local a_i = torch.linspace(-60,60,opt.manif_size)
      for i = 1, opt.manif_size do
        ang:select(2,i):cos((t:select(2,2)-a_i[i])*(math.pi/180))
      end
      --ang:select(2,2):sin(t:select(2,2)*(math.pi/180))
      res[2] = ang
      return res
    end
  end
elseif opt.ptype == 'nopose' then
  bp.setLabel = function(self, dd, do_flip)
    return dd[5]
  end

elseif opt.ptype == 'sep-discrete' then
else
  error('Forgot something')
end

if not opt.only_test then
  bp:setupData()
end

if opt.synth_data ~= -1 and not opt.only_test then
  print('Loading synthetic data from render4cnn')
  local nPoses = opt.ang_res
  local divAng = 360/nPoses
  local tds = require 'tds'
  -- slightly modified to also read the targets
  local function read_render4cnn_data(file)
    -- get all lines from a file, returns an empty 
    -- list/table if the file does not exist
    if not paths.filep(file) then return {} end
    local lines = tds.Vec()--{}
    local targets = tds.Vec()----{}
    for line in io.lines(file) do
      local l = {}
      for i in string.gmatch(line, "%S+") do
        table.insert(l, i)
      end
      local cl_id = tonumber(l[2])+1
      local ang = tonumber(l[1]:sub(-22,-20))
      --table.insert(targets,{tonumber(l[2]), tonumber(l[1]:sub(-22,-20))})
      --table.insert(lines,l[1])
      --table.insert(targets, add_bg + (cl_id-2)*nPoses + math.fmod(math.ceil((ang-divAng/2)/divAng),nPoses)+1)
      lines:insert(l[1])
      targets:insert(add_bg + (cl_id-2)*nPoses + math.fmod(math.ceil((ang-divAng/2)/divAng),nPoses)+1)
    end
    return lines, targets
  end

  local basefolder = '/media/francisco/Seagate Expansion Drive/francisco/work/datasets/render4cnn'

  local imlist, targets = read_render4cnn_data(paths.concat(basefolder,'imglist.txt'))
  local num_imgs = 9
  local old_bp = bp
  
  MixBP = dofile 'experiments/add_render4cnn_data.lua'
  bp = MixBP(old_bp, imlist, targets, basefolder, opt.synth_data)

end

print('Batch Provider:')
print(bp)


if false then
  print('Num bp.bboxes', #bp.bboxes)
  for i = 1, #bp.bboxes do
    if bp.bboxes[i][0] then
      if bp.bboxes[i][0]:numel() == 0 then
        print('ops '.. i)
      end
    else
      print('ops2 '.. i)
    end
  end
  --print(bp.bboxes)
end

--------------------------------------------------------------------------------
-- train
--------------------------------------------------------------------------------
if not  opt.only_test then
trainer = nnf.Trainer(model, criterion, bp)

local num_iter = opt.num_iter/opt.disp_iter
local lr_step = opt.lr_step/opt.disp_iter
local save_step = opt.save_step/opt.disp_iter

local logname = 'loss'
local modelname = 'model'
if opt.finetune then
  logname = logname .. '_finetune'
  modelname = modelname .. '_finetune'
end

local logger = optim.Logger(paths.concat(rundir, logname))
logger:setNames{'loss'}
logger.showPlot = false

trainer.optimState.learningRate = opt.lr

if opt.finetune then
  local lr = opt.lr
  lr = lr/10^math.floor(num_iter/lr_step)
  trainer.optimState.learningRate = lr
  print('Starting with learning rate '..lr)
end

-- main training loop
for i=1,num_iter do
  if i % lr_step == 0 then
    trainer.optimState.learningRate = trainer.optimState.learningRate/10
  end
  print(('Iteration %3d/%-3d'):format(i,num_iter))
  trainer:train(opt.disp_iter)
  print(('  Training error: %.5f'):format(trainer.fx[i]))

  logger:add({trainer.fx[i]})
  logger:style{'-'}
  logger:plot()

  if i% save_step == 0 then
    --torch.save(paths.concat(rundir, 'model.t7'), model:clearState())
    torch.save(paths.concat(rundir, modelname..'.t7'), model)
  end
end

--torch.save(paths.concat(rundir, 'model.t7'), model:clearState())
torch.save(paths.concat(rundir, modelname..'.t7'), model)
end
--------------------------------------------------------------------------------
-- evaluation
--------------------------------------------------------------------------------

optnet.removeOptimization(model)
optnet.optimizeMemory(model,{torch.rand(1,3,224,224):cuda(),torch.CudaTensor{{1,1,1,50,50}}})

-- add softmax to classifier, because we were using nn.CrossEntropyCriterion
local softmax = nn.SoftMax()
softmax:type(tensor_type)

if opt.ptype == 'discrete' or opt.ptype == 'rendercnn' or opt.ptype == 'nopose' 
   or opt.ptype == 'sep-discrete' or opt.ptype == 'discrete-altnorm' then
  model:add(softmax)
elseif opt.ptype == 'mixed' then
  if opt.ang_place == 'fc7' then
    model:get(4):get(7):add(softmax)
  elseif opt.ang_place == 'pool5' then
    model:get(4):get(1):add(softmax)
  elseif opt.ang_place == 'conv4' then
    model:get(3):get(1):add(softmax)
  end
  --[[
  -- clean that up
  local seq = nn.Sequential()
  seq:add(model:get(4):get(7):get(1)) --
  seq:add(softmax)
  local cc = model:get(4):get(7):get(2)
  model:get(4):remove()
  local ccat = nn.ConcatTable()
  ccat:add(seq)
  ccat:add(cc)
  model:get(4):add(ccat)
  --]]
  --model:add(softmax)
end

-- dataset for evaluation
local roidbfile
if opt.testproposals == 'deepmask' then
  roidbfile='data/MAD-DM2k/voc_2012_val.t7'
elseif opt.testproposals == 'selectivesearch' then
  roidbfile='data/selective_search_data/voc_2012_val.mat'
elseif opt.testproposals == 'DMSS' then
  roidbfile='data/MAD-DM2k/voc_2012_val.t7'
else
  error('Unknown testproposals '..opt.testproposals)
end
dsv = nnf.DataSetPascal{
  image_set='val',
  datadir='datasets/VOCdevkit',
  dataset_name='Pascal3D',
  --roidbdir='data/selective_search_data',
  --roidbfile='data/MAD-DM2k/voc_2012_val.t7',
  roidbfile=roidbfile,
  year=2012,
  classes=cls
}

if opt.testproposals == 'DMSS' then
  dsv:loadROIDB()
  print('Concatenating ROIs from selective search to DeepMask')
  local roidb = dsv.roidb
  dsv.roidb = nil
  dsv.roidbfile = 'data/selective_search_data/voc_2012_val.mat'
  dsv:loadROIDB()
  for i = 1, ds:size() do
    dsv.roidb[i] = torch.cat(roidb[i], dsv.roidb[i], 1)
  end
end

print('DataSet Evaluation:')
print(dsv)

-- feature provider for evaluation
fpv = nnf.FRCNN{
  scale={600},
  max_size=1000,
  image_transformer=image_transformer,
}
fpv:evaluate()
print('Feature Provider Evaluation:')
print(fpv)

-- define the class to test the model on the full dataset
tester = nnf.Tester(model, fpv, dsv)
tester.cachefolder = rundir
if opt.finetune then
  tester.cachefolder = paths.concat(rundir,'finetune')
end

tester.nPoses = opt.ang_res

pose_utils = paths.dofile('evaluation_utils.lua')
-- redefining evaluation functions

if opt.ptype == 'discrete' or opt.ptype == 'rendercnn' or opt.ptype == 'discrete-altnorm' then
  tester.finalPrunning = pose_utils.finalPrunning
  tester.pruneDetections = pose_utils.pruneDetections
  tester.evaluate = pose_utils.evaluate

  if opt.sep_class then
    --local cmodel = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/alexnet_DeepMask_VOC2012,model=alexnet/model.t7')
    local cmodel = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/frcnn_DeepMask_flip_nohard,model=alexnet/model.t7')
    cmodel:add(softmax:clone())
    cmodel:evaluate()

    local detec_2 = nnf.ImageDetect(cmodel, fpv)
    tester.frcnn_detec = detec_2
    tester.pruneDetections = pose_utils.pruneDetectionsSeparate
    tester.add_bg = add_bg

    if opt.ptype == 'rendercnn' then
      tester.discretization_step = 360
      tester.discretize = true
    end
  end

elseif opt.ptype == 'mixed' then
  tester.finalPrunning = pose_utils.finalPrunning
  tester.pruneDetections = pose_utils.pruneDetectionsMixed
  tester.evaluate = pose_utils.evaluate
  tester.discretize = true
  tester.manif_size = opt.manif_size
  if opt.sep_class then
    local cmodel = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/frcnn_DeepMask_flip_nohard,model=alexnet/model.t7')
    cmodel:add(softmax:clone())
    cmodel:evaluate()

    local detec_2 = nnf.ImageDetect(cmodel, fpv)
    tester.frcnn_detec = detec_2
    tester.pruneDetections = pose_utils.pruneDetectionsMixedSepClassif
    tester.add_bg = add_bg
  end
elseif opt.ptype == 'regression' then

  local cmodel = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/frcnn_DeepMask_flip_nohard,model=alexnet/model.t7')
  cmodel:add(softmax:clone())
  cmodel:evaluate()

  local detec_2 = nnf.ImageDetect(cmodel, fpv)
  tester.frcnn_detec = detec_2

  tester.finalPrunning = pose_utils.finalPrunning
  tester.pruneDetections = pose_utils.pruneDetectionsRegression
  tester.evaluate = pose_utils.evaluate
  tester.discretize = true
  tester.manif_size = opt.manif_size

elseif opt.ptype == 'nopose' then

elseif opt.ptype == 'sep-discrete' then
  tester.finalPrunning = pose_utils.finalPrunning
  tester.pruneDetections = pose_utils.pruneDetections
  tester.evaluate = pose_utils.evaluate

  local cmodel = torch.load('/home/francisco/work/projects/pose-estimation/cachedir/frcnn_DeepMask_flip_nohard,model=alexnet/model.t7')
  cmodel:add(softmax:clone())
  cmodel:evaluate()

  local detec_2 = nnf.ImageDetect(cmodel, fpv)
  tester.frcnn_detec = detec_2
  tester.pruneDetections = pose_utils.pruneDetectionsSeparate
  tester.add_bg = 1

end

local timer = torch.Timer()

tester:test(opt.num_iter)

print(('Total time for detection: %.2fs'):format(timer:time().real))
